#! /usr/bin/python

""" Main program to synchronize btrfs snapshots.  See README.md.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

if True:  # Headers
    if True:  # imports

        import argparse
        import logging
        import os.path
        import re
        import sys

        import BestDiffs
        import ButterStore
        import S3Store
        import Store

theDebug = False

logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')

theVersionFile = os.path.join(os.path.dirname(__file__), "version.txt")
try:
    with open(theVersionFile) as version:
        theVersion = version.readline()
except IOError:
    print("Missing '%s'" % (theVersionFile))
    theVersion = "<unknown>"

command = argparse.ArgumentParser(
    description="Synchronize two sets of btrfs snapshots.",
    epilog="""
<src>, <dst>:   [file://]/path/to/directory/[snapshot]
                ssh://[user@]host/path/to/directory (Not implemented)
                s3://bucket/prefix/[snapshot]

If only <dst> is supplied, just list available snapshots.  The trailing "/"
on a <src> *is* significant.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
See README.md and LICENSE.txt for more info.
    """,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)

command.add_argument('source', metavar='<src>', nargs='?',  # nargs='+',
                     help='a source of btrfs snapshots')
command.add_argument('dest', metavar='<dst>',
                     help='the btrfs snapshots to be updated')

command.add_argument('-n', '--dry-run', action="store_true",
                     help="display what would be transferred, but don't do it",
                     )
command.add_argument('-d', '--delete', action="store_true",
                     help='delete any snapshots in <dst> that are not in <src> (Not implemented)',
                     )

command.add_argument('-q', '--quiet', action="count", default=0,
                     help="""
                     once: don't display progress.
                     twice: only display error messages""",
                     )
command.add_argument('-l', '--logfile', type=argparse.FileType('w'),
                     help='log debugging information to file',
                     )
command.add_argument('-V', '--version', action="version", version='%(prog)s ' + theVersion,
                     help='display version',
                     )

command.add_argument('--remote-receive', action="store_true",
                     help=argparse.SUPPRESS,
                     )
command.add_argument('--remote-send', action="store_true",
                     help=argparse.SUPPRESS,
                     )
command.add_argument('--remote-list', action="store_true",
                     help=argparse.SUPPRESS,
                     )


def _setupLogging(quietLevel, logFile):
    theDisplayFormat = '%(message)s'
    theDebugDisplayFormat = (
        '%(levelname)7s:'
        '%(filename)s[%(lineno)d] %(funcName)s(): %(message)s'
    )
    theLogFormat = (
        '%(asctime)-15s: %(levelname)7s:'
        '%(filename)s[%(lineno)d] %(funcName)s(): %(message)s'
    )

    root = logging.getLogger()
    root.setLevel("INFO" if theDebug else "DEBUG")  # When debugging, this is handled per-logger

    def add(handler, level, format):
        handler.setLevel(level)
        handler.setFormatter(logging.Formatter(format))
        root.addHandler(handler)

    level = "DEBUG" if theDebug else "INFO" if quietLevel < 2 else "WARN"
    formatString = theDebugDisplayFormat if theDebug else theDisplayFormat

    add(logging.StreamHandler(sys.stdout), level, formatString)

    if logFile is not None:
        add(logging.StreamHandler(logFile), "DEBUG", theLogFormat)


def parseSink(uri, isDest):
    """ Parse command-line description of sink into a sink object. """
    if uri is None:
        return None

    # logger.debug(uri)
    pattern = re.compile('^((?P<method>[^:/]*)://)?(?P<host>[^/]*)(/(?P<path>.*))?$')
    match = pattern.match(uri)
    if match is None:
        # logger.error("Can't parse snapshot store '%s'", uri)
        raise Exception("Can't parse snapshot store '%s'" % (uri))
    parts = match.groupdict()

    if parts['method'] is None:
        parts['method'] = 'file'

    if parts['method'] == 'file':
        parts['path'] = parts['host'] + '/' + parts['path']

    logger.debug(parts)

    Sinks = {
        'file': ButterStore.ButterStore,
        's3': S3Store.S3Store,
        # 'ssh': SSHStore.SSHStore,
    }

    return Sinks[parts['method']](parts['host'], parts['path'], isDest)


def main():
    """ Main program. """
    args = command.parse_args()

    _setupLogging(args.quiet, args.logfile)

    logger.debug("Arguments: %s", vars(args))

    progress = args.quiet == 0

    source = parseSink(args.source, False)

    dest = parseSink(args.dest, source is not None)

    if source is None:
        for vol in dest.listVolumes():
            print vol.display(dest, detail="line")
        return 0

    vols = source.listVolumes()

    best = BestDiffs.BestDiffs(vols, args.delete)
    best.analyze(source, dest)

    summary = best.summary()
    logger.info("Optimal synchronization:")
    for sink, values in summary.items():
        logger.info("%s from %d diffs in %s",
                    Store.humanize(values.size),
                    values.count,
                    sink or "TOTAL",
                    )

    for diff in best.iterDiffs():
        if diff is None:
            raise Exception("Missing diff.  Can't fully replicate.")
            continue

        logger.info("%s: %s", "Keep" if diff.sink == dest else "Xfer", diff)

        vol = diff.toVol
        paths = diff.sink.getPaths(vol)

        if diff.sink != dest:
            streamContext = dest.receive(diff, paths, dryrun=args.dry_run)

            diff.sink.send(diff, streamContext, progress=progress, dryrun=args.dry_run)

            # TODO: For symmetry, put this into the streamContext.__exit__ method

            if vol.hasInfo():
                infoContext = dest.receiveVolumeInfo(paths, dryrun=args.dry_run)

                if args.dry_run:
                    vol.writeInfo(sys.stdout)
                else:
                    with infoContext as stream:
                        vol.writeInfo(stream)

    return 0

if __name__ == "__main__":
    sys.exit(main())



""" Interface to btrfs-tools for snapshots.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

if True:  # Headers
    if True:  # imports
        import os
        import os.path
        import psutil
        import re
        import subprocess
        import sys

        import Store

    if True:  # constants
        import logging
        logger = logging.getLogger(__name__)

        theChunkSize = 100 * (2 ** 20)

        DEVNULL = open(os.devnull, 'wb')

# logger.setLevel('DEBUG')


class Butter:

    """ Interface to local btrfs file system snapshots. """

    def __init__(self):
        """ Initialize. """
        self.btrfsVersion = self._getVersion([3, 14])

    def _getVersion(self, minVersion):
        btrfsVersionString = subprocess.check_output(
            ["btrfs", "--version"], stderr=sys.stderr
        ).decode("utf-8").strip()

        versionPattern = re.compile("[0-9]+(\.[0-9]+)*")
        version = versionPattern.search(btrfsVersionString)

        try:
            version = [int(num) for num in version.group(0).split(".")]
        except AttributeError:
            version = None

        if version < [3, 14]:
            logger.error(
                "%s is not supported.  Please upgrade your btrfs to at least 3.14",
                btrfsVersionString
            )
        else:
            logger.debug("%s", btrfsVersionString)

        return btrfsVersionString

    def receive(self, directory, dryrun=False):
        """ Return a file-like (stream) object to store a diff. """
        cmd = ["btrfs", "receive", directory]

        if Store.skipDryRun(logger, dryrun)("Command: %s", cmd):
            return None

        process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stderr=sys.stderr, stdout=DEVNULL)

        ps = psutil.Process(process.pid)
        ps.ionice(psutil.IOPRIO_CLASS_IDLE)

        return process.stdin

    def send(self, targetPath, parent, streamContext, progress=True, dryrun=False):
        """ Write a (incremental) snapshot to the stream context manager. """
        if parent is not None:
            cmd = ["btrfs", "send", "-p", parent, targetPath]
        else:
            cmd = ["btrfs", "send", targetPath]

        if Store.skipDryRun(logger, dryrun)("Command: %s", cmd):
            return

        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=sys.stderr)
        ps = psutil.Process(process.pid)
        ps.ionice(psutil.IOPRIO_CLASS_IDLE)

        try:
            streamContext.metadata['btrfsVersion'] = self.btrfsVersion
        except AttributeError:
            pass

        try:
            streamContext.progress = progress
        except AttributeError:
            pass

        with streamContext as stream:
            while True:
                data = process.stdout.read(theChunkSize)
                if len(data) == 0:
                    break
                stream.write(data)



""" Utilities for working with fnctl.ioctl. """

# See <linux/ioctl.h> for C source.

import array
import collections
import fcntl
import itertools
import os
import struct

# constant for linux portability
NRBITS = 8
TYPEBITS = 8

# architecture specific
SIZEBITS = 14
DIRBITS = 2

NRMASK = (1 << NRBITS) - 1
TYPEMASK = (1 << TYPEBITS) - 1
SIZEMASK = (1 << SIZEBITS) - 1
DIRMASK = (1 << DIRBITS) - 1

NRSHIFT = 0
TYPESHIFT = NRSHIFT + NRBITS
SIZESHIFT = TYPESHIFT + TYPEBITS
DIRSHIFT = SIZESHIFT + SIZEBITS

NONE = 0
WRITE = 1
READ = 2


class t:

    """ Type definitions for translating linux C headers to Python struct format values. """

    u8 = 'B'
    u16 = 'H'
    u32 = 'L'
    u64 = 'Q'

    (le16, le32, le64) = (u16, u32, u64)  # Works on Linux x86

    char = 'c'

    max_u32 = (1 << 32) - 1
    max_u64 = (1 << 64) - 1

    @staticmethod
    def writeChar(value):
        """ Write a single-character string as a one-byte (u8) number. """
        return 0 if value is None else ord(value[0])

    @staticmethod
    def writeString(data):
        """ Write a string as null-terminated c string (bytes). """
        if data is None:
            return chr(0)

        return data.encode('utf-8') + chr(0)

    @staticmethod
    def readString(data):
        """ Read a null-terminated (c) string. """
        # CAUTION: Great for strings, horrible for buffers!
        return data.decode('utf-8').partition(chr(0))[0]

    @staticmethod
    def readBuffer(data):
        """ Trim zero bytes in buffer. """
        # CAUTION: Great for strings, horrible for buffers!
        return data.rstrip(chr(0))


def unzip(listOfLists):
    """ Inverse of zip to split lists. """
    return zip(*listOfLists)


class _SkipType:

    def popValue(self, argList):
        return None

    def yieldArgs(self, arg):
        if False:
            yield None  # Make this a generator


class _TypeWriter:

    def __init__(self, default, reader=None, writer=None):
        self._default = default
        self._writer = writer or (lambda x: x)
        self._reader = reader or (lambda x: x)

    def popValue(self, argList):
        return self._reader(argList.pop())

    def yieldArgs(self, arg):
        yield self._writer(arg) or self._default


class Structure:

    """ Model a C struct.

    Encapsulates a struct format with named item values.

    structure fields are (typeDef, name, len=1) arguments.
    typeDef can be a string or a structure itself.

    Example Structures:
        >>> s1 = Structure((t.char, 'char1'))
        >>> s2 = Structure(
        ... (t.u16, 'foo'),
        ... (t.u8, 'bar', 8, t.readString, t.writeString),
        ... (s1, 'foobar'),
        ... )
        >>> s2.fmt
        'H8sc'
        >>> s2.size
        11

    Instance variables:
        >>> s2._Tuple(1,2,3).__dict__
        OrderedDict([('foo', 1), ('bar', 2), ('foobar', 3)])
        >>> s2._packed
        True
        >>> s2._types.keys()
        ['foo', 'bar', 'foobar']

    Using a Structure:
        >>> myValues = dict(foo=8, bar=u"hola", foobar=dict(char1='a'))
        >>> data = s2.write(myValues)
        >>> data
        array('B', [8, 0, 104, 111, 108, 97, 0, 0, 0, 0, 97])
        >>> values = s2.read(data)
        >>> values
        StructureTuple(foo=8, bar=u'hola', foobar=StructureTuple(char1='a'))
        >>> values.foo
        8
        >>> values.foobar.char1
        'a'

    """

    def __init__(self, *fields, **keyArgs):
        """ Initialize. """
        (names, formats, types) = unzip([self._parseDefinition(*f) for f in fields])

        self._Tuple = collections.namedtuple("StructureTuple", names)

        self._fmt = "".join(formats)
        self._packed = keyArgs.get('packed', True)
        self._struct = struct.Struct("=" + self._fmt if self._packed else self._fmt)

        self._types = collections.OrderedDict(zip(names, types))

    @property
    def size(self):
        """ Total packed data size. """
        return self._struct.size

    @property
    def fmt(self):
        """ struct module format string without the leading byte-order character. """
        return self._fmt

    # This produces a dictionary of { fmtChar: defaultValue }
    defaults = dict(itertools.chain(*[
        [(fmtChar, defaultValue) for fmtChar in s]
        for (s, defaultValue) in [
            ("sp", ""),
            ("bBhHiIlLqQfdP", 0),
            ("?", False),
            ("c", chr(0)),
        ]]))

    skipType = _SkipType()

    @staticmethod
    def _parseDefinition(typeDef, name, len=1, reader=None, writer=None):
        """ Return (name, format, type) for field.

        type.popValue() and type.yieldArgs() must be implemented.

        """
        if isinstance(typeDef, Structure):
            return (name, typeDef.fmt, typeDef)

        if len != 1:
            size = struct.calcsize(typeDef)
            if typeDef not in "xspP":
                typeDef = 's'
            typeDef = str(len * size) + typeDef

        fmtChar = typeDef[-1:]

        if fmtChar == 'x':
            typeObj = Structure.skipType
        else:
            typeObj = _TypeWriter(Structure.defaults[fmtChar], reader, writer)

        return (name, typeDef, typeObj)

    def yieldArgs(self, keyArgs):
        """ Take (nested) dict(s) of args to set, and return flat list of args. """
        for (name, typeObj) in self._types.items():
            for arg in typeObj.yieldArgs(keyArgs.get(name, None)):
                yield arg

    def write(self, keyArgs):
        """ Write specified key arguments into data structure. """
        # bytearray doesn't work with fcntl
        args = array.array('B', (0,) * self.size)
        self._struct.pack_into(args, 0, *list(self.yieldArgs(keyArgs)))
        return args

    def popValue(self, argList):
        """ Take a flat arglist, and pop relevent values and return as a value or tuple. """
        # return self._Tuple(*[name for (name, typeObj) in self._types.items()])
        return self._Tuple(*[typeObj.popValue(argList) for (name, typeObj) in self._types.items()])

    def read(self, data, offset=0):
        """ Read data structure and return (nested) named tuple(s). """
        args = list(self._struct.unpack_from(data, offset))
        args.reverse()
        return self.popValue(args)


class Buffer:

    """ Contains bytes and an offset. """

    def __init__(self, buf, offset=0):
        """ Initialize. """
        self.buf = buf
        self.offset = offset

    def read(self, structure):
        """ Read and advance. """
        start = self.offset
        self.skip(structure.size)
        return structure.read(self.buf, start)

    def skip(self, len):
        """ Advance. """
        self.offset += len

    def view(self, len):
        """ Return a view of the next len bytes. """
        return memoryview(self.buf)[self.offset:self.offset + len]

    @property
    def len(self):
        """ Count of remaining bytes. """
        return len(self.buf) - self.offset


class Control:

    """ Callable linux io control (ioctl). """

    def __init__(self, direction, op, structure):
        """ Initialize. """
        self.structure = structure
        size = structure.size if structure else 0
        self.ioc = self._iocNumber(direction, self.magic, op, size)

    def __call__(self, device, **args):
        """ Execute the call. """
        if device.fd is None:
            raise Exception("Device hasn't been successfully opened.  Use 'with' statement.")

        if self.structure is not None:
            args = self.structure.write(args)
            # log.write(args)
            ret = fcntl.ioctl(device.fd, self.ioc, args, True)
            # log.write(args)
            assert ret == 0, ret
            return self.structure.read(args)
        else:
            ret = fcntl.ioctl(device.fd, self.ioc)
            assert ret == 0, ret

    @staticmethod
    def _iocNumber(dir, type, nr, size):
        return dir << DIRSHIFT | \
            type << TYPESHIFT | \
            nr << NRSHIFT | \
            size << SIZESHIFT

    @classmethod
    def _IOC(cls, dir, op, structure=None):
        """ Encode an ioctl id. """
        control = cls(dir, op, structure)

        def do(dev, **args):
            return control(dev, **args)
        return do

    @classmethod
    def IO(cls, op):
        """ Returns an ioctl Device method with no arguments. """
        return cls._IOC(NONE, op)

    @classmethod
    def IOW(cls, op, structure):
        """ Returns an ioctl Device method with WRITE arguments. """
        return cls._IOC(WRITE, op, structure)

    @classmethod
    def IOWR(cls, op, structure):
        """ Returns an ioctl Device method with READ and WRITE arguments. """
        return cls._IOC(READ | WRITE, op, structure)

    @classmethod
    def IOR(cls, op, structure):
        """ Returns an ioctl Device method with READ arguments. """
        return cls._IOC(READ, op, structure)


class Device(object):

    """ Context manager for a linux file descriptor for a file or device special file.

    Opening and closing is handled by the Python "with" statement.

    """

    def __init__(self, path, flags=os.O_RDONLY):
        """ Initialize. """
        self.path = path
        self.fd = None
        self.flags = flags

    def __enter__(self):
        """ Open. """
        self.fd = os.open(self.path, self.flags)
        return self

    def __exit__(self, exceptionType, exception, trace):
        """ Close. """
        os.close(self.fd)
        self.fd = None



""" Module to select best snapshot "send" commands to use for synchronization.

Based on optimzing a Directed Acyclic Graph (DAG),
where snapshots are the nodes,
and "send" diffs are the directed edges.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

import Store

import collections
import logging
logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')


class Bunch(object):

    """ Simple mutable data record. """

    def __init__(self, **kwds):
        """ Initialize. """
        self.__dict__.update(kwds)


class _Node:

    def __init__(self, volume, intermediate=False):
        self.volume = volume
        self.intermediate = intermediate
        self.diff = None
        self.height = None

    @property
    def diffSize(self):
        return self.diff.size if self.diff is not None else None

    @property
    def previous(self):
        return self.diff.fromVol if self.diff else None

    @property
    def sink(self):
        return self.diff.sink if self.diff else None

    def __unicode__(self):
        return self.display()

    def display(self, sink=None):
        if self.height is not None:
            ancestors = " (%d ancestors)" % (self.height - 1)
        else:
            ancestors = ""

        return u"%s%s" % (
            self.diff or self.volume.display(sink),
            ancestors,
        )

    def __str__(self):
        return unicode(self).encode('utf-8')

    @staticmethod
    def summary(nodes):
        sinks = collections.defaultdict(lambda: Bunch(count=0, size=0))

        total = sinks[None]

        for n in nodes:
            total.count += 1

            if n.diff is None:
                continue

            total.size += n.diff.size

            sink = sinks[n.diff.sink]
            sink.count += 1
            sink.size += n.diff.size

        return sinks


class BestDiffs:

    """ This analyzes and stores an optimal network (tree).

    The nodes are the desired (or intermediate) volumes.
    The directed edges are diffs from an available sink.

    """

    def __init__(self, volumes, delete=False):
        """ Initialize.

        volumes are the required snapshots.

        """
        self.nodes = {volume: _Node(volume, False) for volume in volumes}
        self.dest = None
        self.delete = delete

    def analyze(self, *sinks):
        """  Figure out the best diffs to use to reach all our required volumes. """
        # Use destination (already uploaded) edges first
        sinks = list(sinks)
        sinks.reverse()
        self.dest = sinks[0]

        nodes = [None]
        height = 1

        def sortKey(node):
            if node is None:
                return None
            return (node.intermediate, self._totalSize(node))

        while len(nodes) > 0:
            logger.debug("Analyzing %d nodes for height %d...", len(nodes), height)

            nodes.sort(key=sortKey)

            for fromNode in nodes:
                if self._height(fromNode) >= height:
                    continue

                if fromNode is not None and fromNode.diffSize is None:
                    continue

                fromSize = self._totalSize(fromNode)
                fromVol = fromNode.volume if fromNode else None

                logger.debug(
                    "Following edges from %s (total %s)",
                    fromNode.display(sinks[-1]) if fromNode is not None else None,
                    Store.humanize(fromSize),
                )

                for sink in sinks:
                    logger.debug(
                        "Listing edges in %s",
                        sink
                    )

                    for edge in sink.getEdges(fromVol):
                        toVol = edge.toVol

                        logger.debug("Edge: %s", edge)

                        # Skip any edges already in the destination
                        if sink != self.dest and self.dest.hasEdge(edge):
                            continue

                        if toVol in self.nodes:
                            toNode = self.nodes[toVol]
                        else:
                            toNode = _Node(toVol, True)
                            self.nodes[toVol] = toNode

                        newCost = self._cost(sink, edge.size, fromSize, height)
                        if toNode.diff is None:
                            oldCost = None
                        else:
                            oldCost = self._cost(
                                toNode.sink,
                                toNode.diffSize,
                                fromSize,
                                self._height(toNode)
                            )

                        # Don't use a more-expensive path
                        if oldCost is not None and oldCost <= newCost:
                            continue

                        # Don't create circular paths
                        if self._wouldLoop(fromVol, toVol):
                            logger.debug("Ignoring looping edge: %s", toVol.display(sink))
                            continue

                        logger.debug(
                            "Replacing edge (%s -> %s cost) %s",
                            Store.humanize(oldCost),
                            Store.humanize(newCost),
                            toNode.display(sink)
                        )

                        toNode.diff = edge

            nodes = [node for node in self.nodes.values() if self._height(node) == height]
            height += 1

        self._prune()

        for node in self.nodes.values():
            node.height = self._height(node)
            if node.diff is None:
                logger.error(
                    "No source diffs for %s",
                    node.volume.display(sinks[-1], detail="line"),
                )

    def _getNode(self, vol):
        return self.nodes[vol] if vol is not None else None

    def _height(self, node):
        if node is None:
            return 0
        else:
            return 1 + self._height(self._getNode(node.previous))

    def _totalSize(self, node):
        if node is None:
            return 0

        prevSize = self._totalSize(self._getNode(node.previous))

        return (node.diffSize or 0) + prevSize

    def _wouldLoop(self, fromVol, toVol):
        if toVol is None:
            return False

        while fromVol is not None:
            if fromVol == toVol:
                return True

            fromVol = self.nodes[fromVol].previous

        return False

    def iterDiffs(self):
        """ Return all diffs used in optimal network. """
        nodes = self.nodes.values()
        nodes.sort(key=lambda node: self._height(node))
        for node in nodes:
            yield node.diff
            # yield { 'from': node.previous, 'to': node.uuid, 'sink': node.diffSink,

    def summary(self):
        """ Return summary count and size in a dictionary. """
        return _Node.summary(self.nodes.values())

    def _prune(self):
        """ Get rid of all intermediate nodes that aren't needed. """
        done = False
        while not done:
            done = True
            for node in [node for node in self.nodes.values() if node.intermediate]:
                if not [dep for dep in self.nodes.values() if dep.previous == node.volume]:
                    logger.debug("Removing unnecessary node %s", node)
                    del self.nodes[node.volume]
                    done = False

    def _cost(self, sink, size, prevSize, height):
        cost = 0

        # Transfer
        if sink != self.dest:
            cost += size

        # Storage
        if sink != self.dest or self.delete:
            cost += size

        # Corruption risk
        cost += (prevSize + size) * (2 ** (height - 8))

        logger.debug("_cost=%d (%s %d %d %d)", cost, sink, size, prevSize, height)

        return cost



""" Back-end for AWS S3.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

# Docs: http://boto.readthedocs.org/en/latest/

from __future__ import division

if True:  # Imports and constants
    if True:  # Imports
        import Store

        import boto
        import collections
        import datetime
        import io
        import logging
        import re
        import sys
    if True:  # Constants
        # For S3 uploads
        theChunkSize = 100 * 2 ** 20

        # Maximum xumber of progress reports per chunk
        theProgressCount = 50

        # This does transparent S3 server-side encryption
        isEncrypted = True

        # Minimum chuck size is 5M.
        theInfoBufferSize = 10 * (1 << 20)

        logger = logging.getLogger(__name__)

# logger.setLevel('DEBUG')


class S3Store(Store.Store):

    """ An S3 bucket synchronization source or sink. """

    def __init__(self, host, path, isDest):
        """ Initialize.

        host is the bucket name.
        path is an object key prefix to use.

        """
        super(S3Store, self).__init__("/" + path, isDest)

        self.isDest = isDest

        self.bucketName = host

        self.keyPattern = re.compile(S3Store.theKeyPattern % ())

        # { fromVol: [diff] }
        self.diffs = None

        logger.info("Listing %s contents...", self)

        try:
            s3 = boto.connect_s3()
        except boto.exception.NoAuthHandlerFound:
            logger.error("Try putting S3 credentials into ~/.boto")
            raise

        self.bucket = s3.get_bucket(self.bucketName)
        self._fillVolumesAndPaths()

    def __unicode__(self):
        """ Return text description. """
        return u'S3 Bucket "%s"' % (self.bucketName)

    def __str__(self):
        """ Return text description. """
        return unicode(self).encode('utf-8')

    def _fillVolumesAndPaths(self):
        """ Fill in self.paths. """
        self.diffs = collections.defaultdict((lambda: []))

        for key in self.bucket.list():
            keyInfo = self._parseKeyName(key.name)

            if keyInfo is None:
                if key.name[-1:] != '/':
                    logger.warn("Ignoring '%s' in S3", key.name)
                continue

            if keyInfo['type'] == 'info':
                stream = io.BytesIO()
                key.get_contents_to_file(stream)
                Store.Volume.readInfo(stream)
                continue

            if keyInfo['from'] == 'None':
                keyInfo['from'] = None

            path = self._relativePath("/" + keyInfo['fullpath'])

            if path is None:
                continue

            diff = Store.Diff(self, keyInfo['to'], keyInfo['from'], key.size)

            logger.debug("Adding %s in %s", diff, path)

            self.diffs[diff.fromVol].append(diff)
            self.paths[diff.toVol].append(path)

        # logger.debug("Diffs:\n%s", pprint.pformat(self.diffs))
        # logger.debug("Vols:\n%s", pprint.pformat(self.vols))

    def getEdges(self, fromVol):
        """ Return the edges available from fromVol. """
        return self.diffs[fromVol]

    def hasEdge(self, diff):
        """ Test whether edge is in this sink. """
        return diff.toVol in [d.toVol for d in self.diffs[diff.fromVol]]

    def receive(self, diff, paths, dryrun=False):
        """ Return Context Manager for a file-like (stream) object to store a diff. """
        path = self.selectReceivePath(paths)
        keyName = self._keyName(diff.toUUID, diff.fromUUID, path)

        if Store.skipDryRun(logger, dryrun)("receive %s in %s", keyName, self):
            return None

        return _Uploader(self.bucket, keyName)

    def receiveVolumeInfo(self, paths, dryrun=False):
        """ Return Context Manager for a file-like (stream) object to store volume info. """
        path = self.selectReceivePath(paths)
        path = path + ".bs"

        if Store.skipDryRun(logger, dryrun)("receive info in '%s'", path):
            return None

        return io.BufferedWriter(_Uploader(self.bucket, path), buffer_size=theInfoBufferSize)

    theKeyPattern = "^(?P<fullpath>.*)/(?P<to>[-a-zA-Z0-9]*)_(?P<from>[-a-zA-Z0-9]*)$"

    def _keyName(self, toUUID, fromUUID, path):
        return "%s/%s_%s" % (path, toUUID, fromUUID)

    def _parseKeyName(self, name):
        """ Returns dict with fullpath, to, from. """
        if name.endswith(".bs"):
            return {'type': 'info'}

        match = self.keyPattern.match(name)
        if not match:
            return None

        match = match.groupdict()
        match.update(type='diff')

        return match

    def send(self, diff, streamContext, progress=True, dryrun=False):
        """ Write the diff (toVol from fromVol) to the stream context manager. """
        path = self.getSendPath(diff.toVol)
        keyName = self._keyName(diff.toUUID, diff.fromUUID, path)
        key = self.bucket.get_key(keyName)

        if Store.skipDryRun(logger, dryrun)("send %s in %s", keyName, self):
            return

        with streamContext as stream:
            if progress:
                key.get_contents_to_file(stream, cb=_DisplayProgress(), num_cb=theProgressCount)
            else:
                key.get_contents_to_file(stream)

        if progress:
            sys.stdout.write("\n")


class _DisplayProgress:

    def __init__(self):
        self.startTime = datetime.datetime.now()

    def __call__(self, sent, total):
        if not sys.stdout.isatty():
            return

        elapsed = datetime.datetime.now() - self.startTime
        mbps = (sent * 8 / (10 ** 6) / elapsed.total_seconds())

        sys.stdout.write(
            "\r %s: Sent %s of %s (%d%%) (%.3g Mbps) %20s\r" % (
                elapsed,
                Store.humanize(sent),
                Store.humanize(total),
                int(100 * sent / total),
                mbps,
                " ",
            )
        )

        sys.stdout.flush()


class _Uploader:

    def __init__(self, bucket, keyName):
        self.bucket = bucket
        self.keyName = keyName
        self.uploader = None
        self.chunkCount = None
        self.metadata = {}
        self.progress = True

    def __enter__(self):
        self.open()
        return self

    def open(self):
        self.uploader = self.bucket.initiate_multipart_upload(
            self.keyName,
            encrypt_key=isEncrypted,
            metadata=self.metadata,
        )
        self.chunkCount = 0

    def __exit__(self, exceptionType, exceptionValue, traceback):
        self.close(abort=exceptionType is not None)
        if exceptionType is not None:
            logger.error("abort")
        else:
            logger.debug("close")
        return False  # Don't supress exception

    def write(self, bytes):
        self.upload(bytes)

    def close(self, abort=False):
        if not abort:
            self.uploader.complete_upload()
            # You cannot change metadata after uploading
            # if self.metadata:
            #     key = self.bucket.get_key(self.keyName)
            #     key.update_metadata(self.metadata)
        else:
            # TODO: this doesn't free storage used by part uploads currently in progress
            self.uploader.cancel_upload()
        self.uploader = None

    def fileno(self):
        raise IOError("S3 uploads don't use file numbers.")

    def upload(self, bytes):
        self.chunkCount += 1
        logger.info(
            "Uploading %s chunk #%d for %s",
            Store.humanize(len(bytes)), self.chunkCount, self.keyName
        )
        fileObject = io.BytesIO(bytes)

        if self.progress:
            self.uploader.upload_part_from_file(
                fileObject, self.chunkCount, cb=_DisplayProgress(), num_cb=theProgressCount
            )
            if sys.stdout.isatty():
                sys.stdout.write("\n")
        else:
            self.uploader.upload_part_from_file(
                fileObject, self.chunkCount
            )



""" Manage read-only volumes in local btrfs file system.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

from __future__ import division

import btrfs
import Butter
import Store

import datetime
import logging
import math
import os
import os.path

logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')
theMinimumChangeRate = .00001


class _Writer:

    """ Context Manager to write a snapshot. """

    def __init__(self, stream, path):
        self.stream = stream
        self.path = path

    def __enter__(self):
        return self.stream.__enter__()

    def __exit__(self, exceptionType, exception, trace):
        self.stream.__exit__(exceptionType, exception, trace)

        if exception is None:
            return

        if not os.path.exists(self.path):
            return

        # This tries to mark partial (failed) transfers.

        partial = self.path + ".part"

        if os.path.exists(partial):
            partial = self.path + "_" + datetime.datetime.now().isoformat() + ".part"

        os.rename(self.path, partial)


class ButterStore(Store.Store):

    """ A local btrfs synchronization source or sink. """

    def __init__(self, host, path, isDest):
        """ Initialize.

        host is ignored.
        path is the file system location of the read-only subvolumes.

        """
        # Don't lose a trailing slash -- it's significant
        path = os.path.abspath(path) + ("/" if path.endswith("/") else "")

        super(ButterStore, self).__init__(path, isDest)

        if not os.path.isdir(self.userPath):
            raise Exception("'%s' is not an existing directory" % (self.userPath))

        self.isDest = isDest

        self.butter = Butter.Butter()  # subprocess command-line interface
        self.btrfs = btrfs.FileSystem(self.userPath)     # ioctl interface

        self.butterVolumes = {}   # Dict of {uuid: <btrfs.Volume>}
        self._fillVolumesAndPaths()

    def _btrfsVol2StoreVol(self, bvol):
        uuid = bvol.received_uuid if self.isDest else bvol.uuid
        if uuid is None:
            return None
        return Store.Volume(uuid, bvol.totalSize, bvol.exclusiveSize, bvol.current_gen)

    def _fillVolumesAndPaths(self):
        """ Fill in self.paths. """
        with self.btrfs as mount:
            for bv in mount.subvolumes:
                if not bv.readOnly:
                    continue

                vol = self._btrfsVol2StoreVol(bv)
                if vol is None:
                    continue

                path = bv.fullPath

                if path is None:
                    logger.warn("Skipping deleted volume %s", bv.uuid)
                    continue

                # vol = Store.Volume(uuid, bv.totalSize, bv.exclusiveSize, bv.gen)
                for path in bv.linuxPaths:
                    path = self._relativePath(path)

                    if path is None:
                        continue

                    logger.debug("%s %s", vol, path)
                    self.paths[vol].append(path)

                if vol not in self.paths:
                    continue

                if vol.uuid in self.butterVolumes:
                    logger.warn(
                        "Duplicate effective uuid %s in '%s' and '%s'",
                        vol.uuid, path, self.butterVolumes[vol.uuid].fullPath
                    )

                self.butterVolumes[vol.uuid] = bv

    def _fileSystemSync(self):
        with self.btrfs as mount:
            mount.SYNC()

    def __unicode__(self):
        """ English description of self. """
        return u"btrfs %s" % (self.userPath)

    def __str__(self):
        """ English description of self. """
        return unicode(self).encode('utf-8')

    def getEdges(self, fromVol):
        """ Return the edges available from fromVol. """
        if fromVol is None:
            for toVol in self.paths:
                yield Store.Diff(self, toVol, fromVol, toVol.size)
            return

        if fromVol not in self.paths:
            return

        fromBVol = self.butterVolumes[fromVol.uuid]
        parentUUID = fromBVol.parent_uuid
        butterDir = os.path.dirname(fromBVol.fullPath)

        vols = [vol for vol in self.butterVolumes.values()
                if vol.parent_uuid == parentUUID or
                os.path.dirname(vol.fullPath) == butterDir
                ]

        changeRate = self._calcChangeRate(vols)

        for toBVol in vols:
            if toBVol == fromBVol:
                continue

            # This gives a conservative estimate of the size of the diff

            estimatedSize = self._estimateSize(toBVol, fromBVol, changeRate)

            toVol = self._btrfsVol2StoreVol(toBVol)

            yield Store.Diff(self, toVol, fromVol, estimatedSize, True)

    def hasEdge(self, diff):
        """ True if Store already contains this edge. """
        return diff.toUUID in self.butterVolumes and diff.fromUUID in self.butterVolumes

    def receive(self, diff, paths, dryrun=False):
        """ Return Context Manager for a file-like (stream) object to store a diff. """
        if not dryrun:
            self._fileSystemSync()

        path = self.selectReceivePath(paths)

        stream = self.butter.receive(os.path.dirname(path), dryrun)

        return _Writer(stream, path) if stream is not None else None

    def receiveVolumeInfo(self, paths, dryrun=False):
        """ Return Context Manager for a file-like (stream) object to store volume info. """
        path = self.selectReceivePath(paths)
        path = path + ".bs"

        if Store.skipDryRun(logger, dryrun)("receive to %s", path):
            return None

        return open(path, "w")

    def _estimateSize(self, toBVol, fromBVol, changeRate):
        fromGen = fromBVol.current_gen
        genDiff = abs(toBVol.current_gen - fromGen)

        estimatedSize = max(0, toBVol.totalSize - fromBVol.totalSize)
        estimatedSize += toBVol.totalSize * (1 - math.exp(-changeRate * genDiff))
        estimatedSize = max(toBVol.exclusiveSize, estimatedSize)

        return estimatedSize

    def _calcChangeRate(self, bvols):
        total = 0
        diffs = 0
        minGen = bvols[0].current_gen
        maxGen = minGen
        minSize = bvols[0].totalSize
        maxSize = minSize

        for vol in bvols:
            total += vol.totalSize
            diffs += vol.exclusiveSize
            minGen = min(minGen, vol.current_gen)
            maxGen = max(maxGen, vol.current_gen)
            minSize = min(minSize, vol.totalSize)
            maxSize = max(maxSize, vol.totalSize)

        try:
            # exclusiveSize is often useless,
            # because data may be shared with read-write volumes not usable for send operations
            diffs = max(diffs, maxSize - minSize)
            rate = - math.log(1 - diffs / total) * (len(bvols) - 1) / (maxGen - minGen)
            rate /= 10  # Fudge
        except (ZeroDivisionError, ValueError):
            logger.debug("Using minimum change rate.")
            rate = theMinimumChangeRate

        logger.debug("Change rate: %f", rate)

        return rate

    def send(self, diff, streamContext, progress=True, dryrun=False):
        """ Write the diff (toVol from fromVol) to the stream context manager. """
        if not dryrun:
            self._fileSystemSync()

        self.butter.send(
            self.getSendPath(diff.toVol),
            self.getSendPath(diff.fromVol),
            streamContext,
            progress,
            dryrun
        )



""" Pythonic wrapper around kernel ioctl-level access to btrfs. """

# See <linux/btrfs.h> for C source.
# See btrfs-progs/ctree.h for some constants.
# Headers taken from 3.13.0-32

# This is NOT a complete implementation,
# just a few useful routines for my current project.

from ioctl import Structure, t
import collections
import ioctl
import logging
import pprint
logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')


def pretty(obj):
    """ Return pretty representation of obj. """
    # if True:
    #     return pprint.pformat(dict(obj.__dict__))
    return pprint.pformat(obj)
    # try:
    #     return pprint.pformat(dict(obj))
    # except TypeError:
    #     try:
    #         return pprint.pformat(obj.__dict__)
    #     except KeyError:
    #         logger.exception("Funny error.")


def bytes2uuid(b):
    """ Return standard human-friendly UUID. """
    if b.strip(chr(0)) == '':
        return None

    s = b.encode('hex')
    return "%s-%s-%s-%s-%s" % (s[0:8], s[8:12], s[12:16], s[16:20], s[20:])


def uuid2bytes(u):
    """ Return compact bytes for UUID. """
    if u is None:
        return ''

    return "".join(u.split('-')).decode('hex')

BTRFS_DEVICE_PATH_NAME_MAX = 1024
BTRFS_SUBVOL_CREATE_ASYNC = (1 << 0)
BTRFS_SUBVOL_RDONLY = (1 << 1)
BTRFS_SUBVOL_QGROUP_INHERIT = (1 << 2)
BTRFS_FSID_SIZE = 16
BTRFS_UUID_SIZE = 16

BTRFS_INO_LOOKUP_PATH_MAX = 4080
btrfs_ioctl_ino_lookup_args = Structure(
    (t.u64, 'treeid'),
    (t.u64, 'objectid'),
    (t.char, 'name', BTRFS_INO_LOOKUP_PATH_MAX, t.readString, t.writeString),
    packed=True
)

btrfs_ioctl_search_key = Structure(
    (t.u64, 'tree_id'),
    (t.u64, 'min_objectid'),
    (t.u64, 'max_objectid'),
    (t.u64, 'min_offset'),
    (t.u64, 'max_offset'),
    (t.u64, 'min_transid'),
    (t.u64, 'max_transid'),
    (t.u32, 'min_type'),
    (t.u32, 'max_type'),
    (t.u32, 'nr_items'),
    (t.u32, 'unused'),
    (t.u64, 'unused1'),
    (t.u64, 'unused2'),
    (t.u64, 'unused3'),
    (t.u64, 'unused4'),
    packed=True
)

btrfs_ioctl_search_header = Structure(
    (t.u64, 'transid'),
    (t.u64, 'objectid'),
    (t.u64, 'offset'),
    (t.u32, 'type'),
    (t.u32, 'len'),
    packed=True
)

btrfs_ioctl_dev_info_args = Structure(
    (t.u64, 'devid'),                # /* in/out */
    (t.u8, 'uuid', BTRFS_UUID_SIZE, bytes2uuid, uuid2bytes),     # /* in/out */
    (t.u64, 'bytes_used'),           # /* out */
    (t.u64, 'total_bytes'),          # /* out */
    (t.u64, 'unused', 379, t.readBuffer),          # /* pad to 4k */
    (t.u8, 'path', BTRFS_DEVICE_PATH_NAME_MAX, t.readString, t.writeString),  # /* out */
    packed=True
)

btrfs_ioctl_fs_info_args = Structure(
    (t.u64, 'max_id'),               # /* out */
    (t.u64, 'num_devices'),          # /* out */
    (t.u8, 'fsid', BTRFS_FSID_SIZE, bytes2uuid, uuid2bytes),     # /* out */
    (t.u64, 'reserved', 124, t.readBuffer),            # /* pad to 1k */
    packed=True
)


BTRFS_SEARCH_ARGS_BUFSIZE = (4096 - btrfs_ioctl_search_key.size)

btrfs_ioctl_search_args = Structure(
    (btrfs_ioctl_search_key, 'key'),
    (t.char, 'buf', BTRFS_SEARCH_ARGS_BUFSIZE),
    packed=True
)

# From btrfs-progs ctree.h

btrfs_disk_key = Structure(
    (t.le64, 'objectid'),
    (t.u8, 'type'),
    (t.le64, 'offset'),
    packed=True
)

btrfs_timespec = Structure(
    (t.le64, 'sec'),
    (t.le32, 'nsec'),
    packed=True
)

btrfs_inode_item = Structure(
    (t.le64, 'generation'),
    (t.le64, 'transid'),
    (t.le64, 'size'),
    (t.le64, 'nbytes'),
    (t.le64, 'block_group'),
    (t.le32, 'nlink'),
    (t.le32, 'uid'),
    (t.le32, 'gid'),
    (t.le32, 'mode'),
    (t.le64, 'rdev'),
    (t.le64, 'flags'),
    (t.le64, 'sequence'),
    (t.le64, 'reserved', 4, t.readBuffer),
    (btrfs_timespec, 'atime'),
    (btrfs_timespec, 'ctime'),
    (btrfs_timespec, 'mtime'),
    (btrfs_timespec, 'otime'),
    packed=True
)

btrfs_root_item_v0 = Structure(
    (btrfs_inode_item, 'inode'),
    (t.le64, 'generation'),
    (t.le64, 'root_dirid'),
    (t.le64, 'bytenr'),
    (t.le64, 'byte_limit'),
    (t.le64, 'bytes_used'),
    (t.le64, 'last_snapshot'),
    (t.le64, 'flags'),
    (t.le32, 'refs'),
    (btrfs_disk_key, 'drop_progress'),
    (t.u8, 'drop_level'),
    (t.u8, 'level'),
    packed=True
)

btrfs_root_item = Structure(
    (btrfs_inode_item, 'inode'),
    (t.le64, 'generation'),
    (t.le64, 'root_dirid'),
    (t.le64, 'bytenr'),
    (t.le64, 'byte_limit'),
    (t.le64, 'bytes_used'),
    (t.le64, 'last_snapshot'),
    (t.le64, 'flags'),
    (t.le32, 'refs'),
    (btrfs_disk_key, 'drop_progress'),
    (t.u8, 'drop_level'),
    (t.u8, 'level'),
    (t.le64, 'generation_v2'),
    (t.u8, 'uuid', BTRFS_UUID_SIZE, bytes2uuid, uuid2bytes),
    (t.u8, 'parent_uuid', BTRFS_UUID_SIZE, bytes2uuid, uuid2bytes),
    (t.u8, 'received_uuid', BTRFS_UUID_SIZE, bytes2uuid, uuid2bytes),
    (t.le64, 'ctransid'),
    (t.le64, 'otransid'),
    (t.le64, 'stransid'),
    (t.le64, 'rtransid'),
    (btrfs_timespec, 'ctime'),
    (btrfs_timespec, 'otime'),
    (btrfs_timespec, 'stime'),
    (btrfs_timespec, 'rtime'),
    (t.le64, 'reserved', 8, t.readBuffer),
    packed=True
)

btrfs_root_ref = Structure(
    (t.le64, 'dirid'),
    (t.le64, 'sequence'),
    (t.le16, 'name_len'),
    packed=True
)

btrfs_qgroup_status_item = Structure(
    (t.le64, 'version'),
    (t.le64, 'generation'),
    (t.le64, 'flags'),
    (t.le64, 'scan'),
    packed=True
)

btrfs_block_group_item = Structure(
    (t.le64, 'used'),
    (t.le64, 'chunk_objectid'),
    (t.le64, 'flags'),
    packed=True
)

btrfs_qgroup_info_item = Structure(
    (t.le64, 'generation'),
    (t.le64, 'referenced'),
    (t.le64, 'referenced_compressed'),
    (t.le64, 'exclusive'),
    (t.le64, 'exclusive_compressed'),
    packed=True
)

btrfs_qgroup_limit_item = Structure(
    (t.le64, 'flags'),
    (t.le64, 'max_referenced'),
    (t.le64, 'max_exclusive'),
    (t.le64, 'rsv_referenced'),
    (t.le64, 'rsv_exclusive'),
    packed=True
)

BTRFS_QUOTA_CTL_ENABLE = 1
BTRFS_QUOTA_CTL_DISABLE = 2
BTRFS_QUOTA_CTL_RESCAN__NOTUSED = 3

btrfs_ioctl_quota_ctl_args = Structure(
    (t.u64, 'cmd'),
    (t.u64, 'status'),
    packed=True
)

btrfs_ioctl_quota_rescan_args = Structure(
    (t.u64, 'flags'),
    (t.u64, 'progress'),
    (t.u64, 'reserved', 6, t.readBuffer),
    packed=True
)

BTRFS_IOCTL_MAGIC = 0x94

objectTypeKeys = {
    'zero': 0,
    'BTRFS_INODE_ITEM_KEY': 1,
    'BTRFS_INODE_REF_KEY': 12,
    'BTRFS_INODE_EXTREF_KEY': 13,
    'BTRFS_XATTR_ITEM_KEY': 24,
    'BTRFS_ORPHAN_ITEM_KEY': 48,
    'BTRFS_DIR_LOG_ITEM_KEY': 60,
    'BTRFS_DIR_LOG_INDEX_KEY': 72,
    'BTRFS_DIR_ITEM_KEY': 84,
    'BTRFS_DIR_INDEX_KEY': 96,
    'BTRFS_EXTENT_DATA_KEY': 108,
    'BTRFS_CSUM_ITEM_KEY': 120,
    'BTRFS_EXTENT_CSUM_KEY': 128,
    'BTRFS_ROOT_ITEM_KEY': 132,
    'BTRFS_ROOT_BACKREF_KEY': 144,
    'BTRFS_ROOT_REF_KEY': 156,
    'BTRFS_EXTENT_ITEM_KEY': 168,
    'BTRFS_METADATA_ITEM_KEY': 169,
    'BTRFS_TREE_BLOCK_REF_KEY': 176,
    'BTRFS_EXTENT_DATA_REF_KEY': 178,
    'BTRFS_EXTENT_REF_V0_KEY': 180,
    'BTRFS_SHARED_BLOCK_REF_KEY': 182,
    'BTRFS_SHARED_DATA_REF_KEY': 184,
    'BTRFS_BLOCK_GROUP_ITEM_KEY': 192,
    'BTRFS_DEV_EXTENT_KEY': 204,
    'BTRFS_DEV_ITEM_KEY': 216,
    'BTRFS_CHUNK_ITEM_KEY': 228,
    'BTRFS_BALANCE_ITEM_KEY': 248,
    'BTRFS_QGROUP_STATUS_KEY': 240,
    'BTRFS_QGROUP_INFO_KEY': 242,
    'BTRFS_QGROUP_LIMIT_KEY': 244,
    'BTRFS_QGROUP_RELATION_KEY': 246,
    'BTRFS_DEV_STATS_KEY': 249,
    'BTRFS_DEV_REPLACE_KEY': 250,
    'BTRFS_UUID_KEY_SUBVOL': 251,
    'BTRFS_UUID_KEY_RECEIVED_SUBVOL': 252,
    'BTRFS_STRING_ITEM_KEY': 253,
}

objectTypeNames = {v: k for (k, v) in objectTypeKeys.iteritems()}


BTRFS_FIRST_FREE_OBJECTID = 256
BTRFS_LAST_FREE_OBJECTID = (1 << 64) - 256
BTRFS_FIRST_CHUNK_TREE_OBJECTID = 256

BTRFS_ROOT_SUBVOL_RDONLY = (1 << 0)

BTRFS_ROOT_TREE_OBJECTID = 1
BTRFS_FS_TREE_OBJECTID = 5
BTRFS_QUOTA_TREE_OBJECTID = 8


class Volume(object):

    """ Represents a subvolume. """

    volumes = {}
    mounts = {}

    def __init__(self, rootid, generation, info):
        """ Initialize. """
        # logger.debug("Volume %d: %s", id, pretty(info))
        self.id = rootid  # id in BTRFS_ROOT_TREE_OBJECTID, also FS treeid for this volume
        self.original_gen = generation
        self.current_gen = info.generation
        # self.size = info.bytes_used
        self.readOnly = bool(info.flags & BTRFS_ROOT_SUBVOL_RDONLY)
        self.level = info.level
        self.uuid = info.uuid
        self.parent_uuid = info.parent_uuid
        self.received_uuid = info.received_uuid

        self.totalSize = None
        self.exclusiveSize = None

        self.info = info

        self.links = {}

        assert rootid not in Volume.volumes, rootid
        Volume.volumes[rootid] = self

        logger.debug("%s", self)

    def _addLink(self, dirTree, dirID, dirSeq, dirPath, name):
        """ Add tree reference and name. (Hardlink). """
        logger.debug("Link  %d-%d-%d '%s%s'", dirTree, dirID, dirSeq, dirPath, name)
        # assert dirTree != 0, (dirTree, dirID, dirSeq, dirPath, name)
        assert (dirTree, dirID, dirSeq) not in self.links, (dirTree, dirID, dirSeq)
        self.links[(dirTree, dirID, dirSeq)] = (dirPath, name)
        assert len(self.links) == 1, self.links  # Cannot have multiple hardlinks to a directory
        logger.debug("%s", self)

    @property
    def fullPath(self):
        """ Return full butter path from butter root. """
        for ((dirTree, dirID, dirSeq), (dirPath, name)) in self.links.items():
            path = Volume.volumes[dirTree].fullPath
            if path is not None:
                return path + ("/" if path[-1] != "/" else "") + dirPath + name

        if self.id == BTRFS_FS_TREE_OBJECTID:
            return "/"
        else:
            return None

    @property
    def linuxPaths(self):
        """ Return full paths from linux root.

        The first path returned will be the path through the top-most mount.
        (Usually the root).
        """
        for ((dirTree, dirID, dirSeq), (dirPath, name)) in self.links.items():
            for path in Volume.volumes[dirTree].linuxPaths:
                yield path + "/" + dirPath + name
        if self.fullPath in Volume.mounts:
            yield Volume.mounts[self.fullPath]

    def __str__(self):
        """ String representation. """
        # logger.debug("%d %d %d", self.gen, self.info.generation, self.info.inode.generation)
        # logger.debug("%o %o", self.info.flags, self.info.inode.flags)
        return """%4d '%s' (level:%d gen:%d total:%d exclusive:%d%s)
        %s (parent:%s received:%s)
        %s%s""" % (
            self.id or -1,
            # ", ".join([dirPath + name for (dirPath, name) in self.links.values()]),
            self.fullPath,
            self.level or -1,
            self.current_gen or -1,
            # self.size,
            self.totalSize or -1,
            self.exclusiveSize or -1,
            " ro" if self.readOnly else "",
            self.uuid,
            self.parent_uuid,
            self.received_uuid,
            "\n\t".join(self.linuxPaths),
            # "\n\t" + pretty(self.__dict__),
            "",
        )


class Control(ioctl.Control):

    """ A btrfs IOCTL. """

    magic = BTRFS_IOCTL_MAGIC


class FileSystem(ioctl.Device):

    """ Mounted file system descriptor for ioctl actions. """

    def __init__(self, path):
        """ Initialize. """
        super(FileSystem, self).__init__(path)

        self.volumes = None
        self.mounts = None

    @property
    def subvolumes(self):
        """ Subvolumes contained in this mount. """
        self.SYNC()
        self._getRoots()
        self._getMounts()
        self._getUsage()

        volumes = Volume.volumes.values()
        volumes.sort(key=(lambda v: v.fullPath))
        return volumes

    def rescanSizes(self):
        """ Zero and recalculate quota sizes to subvolume sizes will be correct. """
        self.QUOTA_CTL(cmd=BTRFS_QUOTA_CTL_ENABLE)
        self.QUOTA_RESCAN()
        logger.info("Waiting for btrfs rescan...")
        self.QUOTA_RESCAN_WAIT()

    def _getMounts(self):
        Volume.mounts = {}
        defaultSubvol = self.DEFAULT_SUBVOL().id
        if defaultSubvol == 0:
            defaultSubvol = ""
        else:
            defaultSubvol = Volume.volumes[defaultSubvol].fullPath.rstrip("/")
        with open("/etc/mtab") as mtab:
            for line in mtab:
                (dev, path, fs, opts, freq, passNum) = line.split()
                if fs != "btrfs":
                    continue
                opts = {
                    opt: value
                    for (opt, _, value) in
                    [o.partition("=") for o in opts.split(",")]
                }
                subvol = "/" + opts.get('subvol', defaultSubvol)
                Volume.mounts[subvol] = path
                logger.debug("%s: %s", subvol, path)

    Key = collections.namedtuple('Key', ('objectid', 'type', 'offset'))

    Key.first = Key(
        objectid=0,
        type=0,
        offset=0,
    )

    Key.last = Key(
        objectid=BTRFS_LAST_FREE_OBJECTID,
        type=t.max_u32,
        offset=t.max_u64,
    )

    Key.next = (lambda key: FileSystem.Key(key.objectid, key.type, key.offset + 1))

    def _walkTree(self, treeid):
        key = FileSystem.Key.first

        while True:
            # Returned objects seem to be monotonically increasing in (objectid, type, offset)
            # min and max values are *not* filters.
            result = self.TREE_SEARCH(
                key=dict(
                    tree_id=treeid,
                    min_type=key.type,
                    max_type=FileSystem.Key.last.type,
                    min_objectid=key.objectid,
                    max_objectid=FileSystem.Key.last.objectid,
                    min_offset=key.offset,
                    max_offset=FileSystem.Key.last.offset,
                    min_transid=0,
                    max_transid=t.max_u64,
                    nr_items=4096,
                ),
            )
            # logger.debug("Search key result: \n%s", pretty(result.key))

            buf = ioctl.Buffer(result.buf)
            results = result.key.nr_items

            logger.debug("Reading %d nodes from %d bytes", results, buf.len)
            if results == 0:
                break

            for _ in xrange(results):
                assert buf.len >= btrfs_ioctl_search_header.size, buf.len
                data = buf.read(btrfs_ioctl_search_header)

                # logger.debug("Object %d: %s", i, pretty(data))

                assert buf.len >= data.len, (buf.len, data.len)
                yield (data, buf.view(data.len))

                key = FileSystem.Key(data.objectid, data.type, data.offset).next()
                buf.skip(data.len)

    def _getRoots(self):
        for (header, buf) in self._walkTree(BTRFS_ROOT_TREE_OBJECTID):
            if header.type == objectTypeKeys['BTRFS_ROOT_BACKREF_KEY']:
                data = Structure(
                    (btrfs_root_ref, 'ref'),
                    (t.char, 'name', header.len - btrfs_root_ref.size),
                ).read(buf)

                info = data.ref
                name = data.name

                directory = self.INO_LOOKUP(treeid=header.offset, objectid=info.dirid)

                logger.debug("%s: %s %s", name, pretty(info), pretty(directory))

                Volume.volumes[header.objectid]._addLink(
                    header.offset,
                    info.dirid,
                    info.sequence,
                    directory.name,
                    name,
                )
            elif header.type == objectTypeKeys['BTRFS_ROOT_ITEM_KEY']:
                if header.len == btrfs_root_item.size:
                    info = btrfs_root_item.read(buf)
                elif header.len == btrfs_root_item_v0.size:
                    info = btrfs_root_item_v0.read(buf)
                else:
                    assert False, header.len

                if ((header.objectid >= BTRFS_FIRST_FREE_OBJECTID and
                        header.objectid <= BTRFS_LAST_FREE_OBJECTID
                     ) or
                        header.objectid == BTRFS_FS_TREE_OBJECTID):
                    Volume(
                        header.objectid,
                        header.offset,
                        info,
                    )

    def _getUsage(self):
        for (header, buf) in self._walkTree(BTRFS_QUOTA_TREE_OBJECTID):
            # logger.debug("%s %s", objectTypeNames[header.type], header)

            if header.type == objectTypeKeys['BTRFS_QGROUP_INFO_KEY']:
                data = btrfs_qgroup_info_item.read(buf)
                try:
                    vol = Volume.volumes[header.offset]
                    vol.totalSize = data.referenced
                    vol.exclusiveSize = data.exclusive
                except KeyError:
                    pass

            elif header.type == objectTypeKeys['BTRFS_QGROUP_LIMIT_KEY']:
                data = btrfs_qgroup_limit_item.read(buf)
            elif header.type == objectTypeKeys['BTRFS_QGROUP_RELATION_KEY']:
                data = None
            else:
                data = None

            # logger.debug('%s', pretty(data))

    def _getDevInfo(self):
        return self.DEV_INFO(devid=1, uuid="")

    def _getFSInfo(self):
        return self.FS_INFO()

    volid_struct = Structure(
        (t.u64, 'id')
    )

    SYNC = Control.IO(8)
    TREE_SEARCH = Control.IOWR(17, btrfs_ioctl_search_args)
    INO_LOOKUP = Control.IOWR(18, btrfs_ioctl_ino_lookup_args)
    DEFAULT_SUBVOL = Control.IOW(19, volid_struct)
    DEV_INFO = Control.IOWR(30, btrfs_ioctl_dev_info_args)
    FS_INFO = Control.IOR(31, btrfs_ioctl_fs_info_args)
    QUOTA_RESCAN = Control.IOW(44, btrfs_ioctl_quota_rescan_args)
    QUOTA_RESCAN_WAIT = Control.IO(46)
    QUOTA_CTL = Control.IOWR(40, btrfs_ioctl_quota_ctl_args)

# define BTRFS_IOC_DEFAULT_SUBVOL _IOW(BTRFS_IOCTL_MAGIC, 19, __u64)
# define BTRFS_IOC_INO_PATHS _IOWR(BTRFS_IOCTL_MAGIC, 35, \
#                     struct btrfs_ioctl_ino_path_args)
# define BTRFS_IOC_LOGICAL_INO _IOWR(BTRFS_IOCTL_MAGIC, 36, \
#                     struct btrfs_ioctl_ino_path_args)
# define BTRFS_IOC_SET_RECEIVED_SUBVOL _IOWR(BTRFS_IOCTL_MAGIC, 37, \
#                 struct btrfs_ioctl_received_subvol_args)
# define BTRFS_IOC_SEND _IOW(BTRFS_IOCTL_MAGIC, 38, struct btrfs_ioctl_send_args)
# define BTRFS_IOC_QUOTA_RESCAN_STATUS _IOR(BTRFS_IOCTL_MAGIC, 45, \
#                    struct btrfs_ioctl_quota_rescan_args)



#! /usr/bin/python

""" Utility program to list data about btrfs subvolumes.

Gives more information than "btrfs sub list", but doesn't filter or format.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
"""

if True:  # imports

    import argparse
    import logging
    import pprint
    import sys

    import btrfs


theDisplayFormat = '%(message)s'
theDebugDisplayFormat = (
    '%(levelname)7s:'
    '%(filename)s[%(lineno)d] %(funcName)s(): %(message)s'
)

logging.basicConfig(format=theDebugDisplayFormat, level='INFO')
logger = logging.getLogger(__name__)
# logger.setLevel('DEBUG')

command = argparse.ArgumentParser(
    description="List data about btrfs subvolumes.",
    epilog="""

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.
See README.md and LICENSE.txt for more info.
    """,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)

command.add_argument('dir', metavar='<dir>',
                     help='list subvolumes in this directory')


def main():
    """ Main program. """
    args = command.parse_args()

    with btrfs.FileSystem(args.dir) as mount:
        # mount.rescanSizes()

        fInfo = mount.FS_INFO()
        pprint.pprint(fInfo)

        dInfo = mount.DEV_INFO(devid=1, uuid="")
        pprint.pprint(dInfo)

        for vol in mount.subvolumes:
            print vol

    return 0

if __name__ == "__main__":
    sys.exit(main())



""" Abstract and component classes for sources and sinks.

Copyright (c) 2014 Ames Cornish.  All rights reserved.  Licensed under GPLv3.

"""

from __future__ import division

import collections
import logging
import math
import os.path

logger = logging.getLogger('__name__')
# logger.setLevel('DEBUG')


class Store(object):

    """ Abstract class for storage back-end.

    Diffs should be indexed by "from" volume.
    Paths should be indexed by volume.

    """

    ignoreExtraVolumes = False

    def __init__(self, userPath, isDest):
        """ Initialize. """
        # { vol: [path] }
        # The order of the paths is important to work around btrfs bugs.
        # The first path is usually the root-volume mounted path,
        # which is required by btrfs 3.14.2.
        self.paths = collections.defaultdict((lambda: []))

        # Paths specify a directory containing subvolumes,
        # unless it's a source path not ending in "/",
        # then it's a single source subvolume.

        if not (userPath.endswith("/") or isDest):
            self.userVolume = os.path.basename(userPath)
            userPath = os.path.dirname(userPath)
        else:
            self.userVolume = None

        # This will not end with a "/"
        userPath = os.path.normpath(userPath)

        assert userPath.startswith("/"), userPath
        self.userPath = userPath

        logger.debug("%s('%s')", self.__class__.__name__, userPath)

    def listVolumes(self):
        """ Return list of all volumes in this Store's selected directory. """
        for (vol, paths) in self.paths.items():
            for path in paths:
                if path.startswith('/'):
                    continue
                if self.userVolume is not None and os.path.basename(path) != self.userVolume:
                    continue
                yield vol
                break

    def getPaths(self, volume):
        """ Return list of all paths to this volume in this Store. """
        return self.paths[volume]

    def getSendPath(self, volume):
        """ Get a path appropriate for sending the volume from this Store.

        The path may be relative or absolute in this Store.

        """
        try:
            return self._fullPath(next(iter(self.getPaths(volume))))
        except StopIteration:
            return None

    def selectReceivePath(self, paths):
        """ From a set of destination paths, select the best one to receive to.

        The paths are relative or absolute, in a source Store.
        The result will be absolute, suitable for this destination Store.

        """
        logger.debug("%s", paths)
        try:
            return self._fullPath([p for p in paths if not p.startswith("/")][0])
        except IndexError:
            return self._fullPath(os.path.basename(list(paths)[0]))

    def _fullPath(self, path):
        if path.startswith("/"):
            return path
        if path == ".":
            return self.userPath
        return os.path.normpath(os.path.join(self.userPath, path))

    def _relativePath(self, fullPath):
        if fullPath is None:
            return None

        assert fullPath.startswith("/"), fullPath

        path = os.path.relpath(fullPath, self.userPath)

        if not path.startswith("../"):
            return path
        elif self.ignoreExtraVolumes:
            return None
        else:
            return fullPath

    # Abstract methods

    def _fillVolumesAndPaths(self):
        """ Fill in self.paths. """
        raise NotImplementedError

    def getEdges(self, fromVol):
        """ Return the edges available from fromVol. """
        raise NotImplementedError

    def hasEdge(self, diff):
        """ True if Store already contains this edge. """
        raise NotImplementedError

    def receive(self, diff, paths, dryrun=False):
        """ Return Context Manager for a file-like (stream) object to store a diff. """
        raise NotImplementedError

    def send(self, diff, streamContext, progress=True, dryrun=False):
        """ Write the diff (toVol from fromVol) to the stream context manager. """
        raise NotImplementedError

    def receiveVolumeInfo(self, paths, dryrun=False):
        """ Return Context Manager for a file-like (stream) object to store volume info. """
        raise NotImplementedError


class Diff:

    """ Represents a btrfs send diff that creates toVol from fromVol. """

    def __init__(self, sink, toVol, fromVol, size, sizeIsEstimated=False):
        """ Initialize. """
        self.sink = sink
        self.toVol = Volume.make(toVol)
        self.fromVol = Volume.make(fromVol)
        self._size = size
        self._sizeIsEstimated = sizeIsEstimated

        if self.fromVol is not None and size is not None and not sizeIsEstimated:
            Diff.theKnownSizes[self.toUUID][self.fromUUID] = size

    # {toVolume: {fromVolume: size}}
    theKnownSizes = collections.defaultdict(lambda: collections.defaultdict(lambda: None))

    @property
    def toUUID(self):
        """ 'to' volume's UUID. """
        return self.toVol.uuid

    @property
    def fromUUID(self):
        """ 'from' volume's UUID, if any. """
        return self.fromVol.uuid if self.fromVol else None

    @property
    def size(self):
        """ Return size. """
        self._updateSize()
        return self._size

    @property
    def sizeIsEstimated(self):
        """ Return whether size is estimated. """
        self._updateSize()
        return self._sizeIsEstimated

    def _updateSize(self):
        if self._size and not self._sizeIsEstimated:
            return

        size = Diff.theKnownSizes[self.toUUID][self.fromUUID]

        if size is None:
            return

        self._size = size
        self._sizeIsEstimated = False

    def __str__(self):
        """ human-readable string. """
        return u"%s from %s (%s%s) in %s" % (
            self.toVol.display(self.sink),
            self.fromVol.display(self.sink) if self.fromVol else "<None>",
            "~" if self.sizeIsEstimated else "",
            humanize(self.size),
            self.sink,
        )


class Volume:

    """ Represents a snapshot. """

    def __init__(self, uuid, size=None, exclusiveSize=None, gen=None):
        """ Initialize. """
        assert uuid is not None
        self._uuid = uuid  # Must never change!
        self.size = size
        self.exclusiveSize = exclusiveSize
        self.gen = gen

    def __cmp__(self, vol):
        """ Compare. """
        return cmp(self._uuid, vol._uuid) if vol else 1

    def __hash__(self):
        """ Hash. """
        return hash(self._uuid)

    @property
    def uuid(self):
        """ Read-only uuid. """
        return self._uuid

    def writeInfo(self, stream):
        """ Write information about diffs into a file stream for use later. """
        for (fromUUID, size) in Diff.theKnownSizes[self.uuid].iteritems():
            if size is not None and fromUUID is not None:
                stream.write(str("%s\t%s\t%d\n" % (
                    self.uuid,
                    fromUUID,
                    size,
                )))

    def hasInfo(self):
        """ Will have information to write. """
        count = len([None
                     for (fromUUID, size)
                     in Diff.theKnownSizes[self.uuid].iteritems()
                     if size is not None and fromUUID is not None
                     ])
        return count > 0

    @staticmethod
    def readInfo(stream):
        """ Read previously-written information about diffs. """
        for line in stream:
            (toUUID, fromUUID, size) = line.split()
            Diff.theKnownSizes[toUUID][fromUUID] = size

    def __unicode__(self):
        """ Friendly string for volume. """
        return self.display()

    def __str__(self):
        """ Friendly string for volume. """
        return unicode(self).encode('utf-8')

    def __repr__(self):
        """ Python expression to create self. """
        return "%s(%s)" % (
            self.__class__,
            self.__dict__,
        )

    def display(self, sink=None, detail='phrase'):
        """ Friendly string for volume, using sink paths. """
        if detail in ('line', 'paragraph') and self.size is not None:
            size = " (%s%s)" % (
                printBytes(self.size),
                "" if self.exclusiveSize is None else (
                    " %s exclusive" % (printBytes(self.exclusiveSize))
                )
            )
        else:
            size = ""

        vol = "%s %s" % (
            printUUID(self._uuid),
            sink.getSendPath(self) if sink else "",
        )

        return vol + size

    @classmethod
    def make(cls, vol):
        """ Convert uuid to Volume, if necessary. """
        if isinstance(vol, cls):
            return vol
        elif vol is None:
            return None
        else:
            return cls(vol)


def printBytes(number):
    """ Return a human-readable string for number. """
    return humanize(number)


def humanize(number):
    """ Return a human-readable string for number. """
    # units = ('bytes', 'KB', 'MB', 'GB', 'TB')
    # base = 1000
    units = ('bytes', 'KiB', 'MiB', 'GiB', 'TiB')
    base = 1024
    if number is None:
        return None
    pow = int(math.log(number, base)) if number > 0 else 0
    pow = min(pow, len(units) - 1)
    mantissa = number / (base ** pow)
    return "%.4g %s" % (mantissa, units[pow])


def printUUID(uuid):
    """ Return friendly abbreviated string for uuid. """
    if uuid is None:
        return None
    return "%s...%s" % (uuid[:4], uuid[-4:])


def skipDryRun(logger, dryRun):
    """ Print or log command about to be run, and return True if should be skipped. """
    return _skipRun if dryRun else logger.debug


def _skipRun(format, *args):
    print("WOULD: " + format % args)
    return True
